# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.object_utils.json_conversion."""

import abc
import typing
import unittest
from pyglove.core.object_utils import json_conversion
from pyglove.core.typing import inspect as pg_inspect


class X:
  """Classes with nested classes for testing purpose."""

  class Y:

    class Z:

      @classmethod
      def class_method(cls):
        return cls.__name__

      @staticmethod
      def static_method():
        return 1

      def instance_method(self):
        return str(self)

  def __init__(self, x):
    self.x = x

  def __eq__(self, other):
    return isinstance(other, X) and self.x == other.x

  def __ne__(self, other):
    return not self.__eq__(other)


def bar():
  pass

T1 = typing.TypeVar('T1')
T2 = typing.TypeVar('T2')
T3 = typing.TypeVar('T3')
T4 = typing.TypeVar('T4')
T5 = typing.TypeVar('T5')


class G1(typing.Generic[T1]):
  pass


class G2(typing.Generic[T1, T2]):
  pass


class G3(typing.Generic[T1, T2, T3]):
  pass


class G4(typing.Generic[T1, T2, T3, T4]):
  pass


class G5(typing.Generic[T1, T2, T3, T4, T5]):
  pass


class JSONConvertibleTest(unittest.TestCase):
  """Tests for JSONConvertible type registry."""

  def test_registry(self):

    class A(json_conversion.JSONConvertible):

      @abc.abstractmethod
      def value(self):
        pass

    class B(A):

      def __init__(self, x):
        super().__init__()
        self.x = x

      def to_json(self):
        return A.to_json_dict({
            'x': self.x
        })

      def value(self):
        return self.x

    typename = lambda cls: f'{cls.__module__}.{cls.__qualname__}'

    # A is abstract.
    self.assertFalse(json_conversion.JSONConvertible.is_registered(typename(A)))
    self.assertTrue(json_conversion.JSONConvertible.is_registered(typename(B)))
    self.assertIs(
        json_conversion.JSONConvertible.class_from_typename(typename(B)), B)
    self.assertIn(
        (typename(B), B),
        list(json_conversion.JSONConvertible.registered_types()))

    class C(B):
      auto_register = False

    # Auto-register is off.
    self.assertFalse(json_conversion.JSONConvertible.is_registered(typename(C)))

    with self.assertRaisesRegex(
        KeyError, 'Type .* has already been registered with class .*'):
      json_conversion.JSONConvertible.register(typename(B), C)

    json_conversion.JSONConvertible.register(
        typename(B), C, override_existing=True)
    self.assertIn(
        (typename(B), C),
        list(json_conversion.JSONConvertible.registered_types()))

  def test_json_conversion(self):

    class T(json_conversion.JSONConvertible):

      def __init__(self, x=None):
        self.x = x

      def to_json(self):
        return T.to_json_dict(dict(x=(self.x, None)), exclude_default=True)

      def __eq__(self, other):
        return isinstance(other, T) and self.x == other.x

      def __ne__(self, other):
        return not self.__eq__(other)

    typename = lambda cls: f'{cls.__module__}.{cls.__qualname__}'
    json_value = json_conversion.to_json([(T(1), 2), {'y': T(3)}])
    self.assertEqual(json_value, [
        ['__tuple__', {'_type': typename(T), 'x': 1}, 2],
        {'y': {'_type': typename(T), 'x': 3}}
    ])
    self.assertEqual(json_conversion.from_json(json_value),
                     [(T(1), 2), {'y': T(3)}])

    # Omitting default values.
    json_value = json_conversion.to_json([(T(), 2), {'y': T(3)}])
    self.assertEqual(json_value, [
        ['__tuple__', {'_type': typename(T)}, 2],
        {'y': {'_type': typename(T), 'x': 3}}
    ])
    self.assertEqual(json_conversion.from_json(json_value),
                     [(T(), 2), {'y': T(3)}])

    # Test bad cases.
    with self.assertRaisesRegex(
        ValueError, 'Tuple should have at least one element besides .*'):
      json_conversion.from_json(['__tuple__'])

    with self.assertRaisesRegex(
        TypeError, 'Type name .* is not registered'):
      json_conversion.from_json({'_type': '__main__.ABC'})

  def assert_conversion_is(self, v):
    self.assertIs(json_conversion.from_json(json_conversion.to_json(v)), v)

  def assert_conversion_equal(self, v):
    self.assertEqual(json_conversion.from_json(json_conversion.to_json(v)), v)

  def test_json_conversion_for_types(self):
    # Built-in types.
    self.assert_conversion_is(int)
    self.assert_conversion_is(bool)
    self.assert_conversion_is(float)
    self.assert_conversion_is(str)
    self.assert_conversion_is(list)
    self.assert_conversion_equal(list[int])
    self.assert_conversion_is(tuple)
    self.assert_conversion_equal(tuple[int, int])
    self.assert_conversion_is(dict)
    self.assert_conversion_equal(dict[str, int])
    self.assert_conversion_is(Exception)
    self.assert_conversion_is(type(None))
    self.assert_conversion_is(...)
    self.assert_conversion_equal(typing.Callable[[int, int], None])
    self.assert_conversion_equal(typing.Callable[..., None])

    # User types.
    self.assert_conversion_is(X)
    self.assert_conversion_is(X.Y)
    self.assert_conversion_is(X.Y.Z)

    # Local type.
    class B:
      pass

    with self.assertRaisesRegex(
        ValueError, 'Cannot convert local class .* to JSON.'):
      json_conversion.to_json(B)

    # Generic types.
    self.assert_conversion_is(G1[int])
    self.assert_conversion_is(G2[int, int])
    self.assert_conversion_is(G3[int, int, int])
    self.assert_conversion_is(G4[int, int, int, int])
    with self.assertRaisesRegex(
        NotImplementedError,
        'Cannot convert generic type with more than 4 type arguments'):
      json_conversion.to_json(G5[int, int, int, int, int])

  def test_json_conversion_for_annotations(self):
    self.assert_conversion_is(typing.Any)
    self.assert_conversion_is(typing.List)
    self.assert_conversion_is(typing.List[typing.List[int]])
    self.assert_conversion_is(typing.Annotated[int, 'abc'])
    self.assert_conversion_is(typing.Dict[str, typing.Any])
    self.assert_conversion_is(typing.Union[int, str])
    self.assert_conversion_is(typing.Sequence[int])
    self.assert_conversion_is(typing.Set[int])
    self.assert_conversion_is(typing.FrozenSet[int])
    self.assert_conversion_is(typing.Mapping[int, str])
    self.assert_conversion_is(typing.MutableMapping[int, str])
    # Optional will be converted to Union[int, None]
    self.assert_conversion_equal(typing.Optional[int])

    with self.assertRaisesRegex(ValueError, 'Annotation cannot be converted'):
      json_conversion.to_json(typing.Literal)

  def test_json_conversion_for_functions(self):
    # Built-in functions.
    self.assert_conversion_is(print)
    self.assert_conversion_is(zip)
    self.assert_conversion_is(input)

    # User-defined functions.
    self.assert_conversion_is(bar)

    # Lambda function.
    s = lambda x, y=1: x + y
    s1 = json_conversion.from_json(json_conversion.to_json(s))
    self.assertTrue(pg_inspect.callable_eq(s1, s))
    self.assertEqual(s1(1), 2)
    self.assertEqual(s1(1, 2), 3)

    # Locally defined function.
    def baz(x, y=1):
      return x + y
    baz1 = json_conversion.from_json(json_conversion.to_json(baz))
    self.assertTrue(pg_inspect.callable_eq(baz1, baz))
    self.assertEqual(baz1(1), 2)
    self.assertEqual(baz1(1, 2), 3)

  def test_json_conversion_for_methods(self):
    # Test class-level method.
    f = json_conversion.from_json(json_conversion.to_json(X.Y.Z.class_method))
    self.assertEqual(f(), 'Z')

    # Test static method.
    self.assert_conversion_is(X.Y.Z.static_method)

    with self.assertRaisesRegex(
        ValueError, 'Cannot convert instance method .* to JSON.'):
      json_conversion.to_json(X.Y.Z().instance_method)

  def test_json_conversion_for_opaque_objects(self):
    self.assert_conversion_equal(X(1))

    class LocalX:
      pass

    with self.assertRaisesRegex(
        ValueError, 'Cannot encode opaque object .* with pickle.'):
      json_conversion.to_json(LocalX())

    json_dict = json_conversion.to_json(X(1))
    json_dict['value'] = 'abc'
    with self.assertRaisesRegex(
        ValueError, 'Cannot decode opaque object with pickle.'):
      json_conversion.from_json(json_dict)

  def test_json_conversion_force_dict(self):
    self.assertEqual(
        json_conversion.from_json([
            '__tuple__',
            1,
            {
                '_type': 'Unknown type',
                'x': [{
                    '_type': 'Unknown type',
                }]
            }
        ], force_dict=True),
        (1, {'x': [{}]})
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic reference."""

import numbers
from typing import Any, Callable, Optional, Tuple
from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import base
from pyglove.core.symbolic.object import Object


class Ref(Object, base.Inferential):
  """Symbolic reference.

  When adding a symbolic node to a symbolic tree, it undergoes a copy operation
  if it already has a parent, ensuring that all symbolic objects have a single
  parent. Additionally, list and dict objects are automatically converted to
  ``pg.List`` and ``pg.Dict``, respectively, to enable symbolic operability.

  However, these two conventions come with certain costs. The act of making
  copies incurs a runtime cost, and it also introduces challenges in sharing
  states across different symbolic objects. To address this issue, symbolic
  reference is introduced. This feature allows a symbolic node to refer to
  value objects without the need for transformation or copying, even when the
  symbolic node itself is copied. For example::

    class A(pg.Object):
    x: int

    a = pg.Ref(A(1))
    b = pg.Dict(x=a)
    c = pg.Dict(y=a)

    assert b.x is a
    assert c.y is a
    assert b.clone().x is a
    assert c.clone(deep=True).y is a

  In this example, ``pg.Ref`` is used to create a symbolic reference to the
  object ``A(1)``, and the ``pg.Dict`` objects `b` and `c` can then reference
  `a` without creating additional copies. This mechanism not only mitigates
  the runtime cost but also facilitates seamless sharing of states among various
  symbolic objects.

  Another useful scenario arises when we wish to utilize regular Python list
  and dict objects. In this case, ``pg.Ref`` enables us to access the list/dict
  object as fields in the symbolic tree without requiring them to be transformed
  into ``pg.List`` and ``pg.Dict``. This allows for seamless integration of
  standard Python containers within the symbolic structure::

    d = pg.Dict(x=pg.Ref({1: 2}))
    assert isinstance(d.x, dict)
    assert not isinstance(d.x, pg.Dict)

    e = pg.Dict(x=pg.Ref([0, 1, 2]]))
    assert isinstance(e.x, list)
    assert not isinstance(e.x, pg.List)

  Please be aware that ``pg.Ref`` objects are treated as leaf nodes in the
  symbolic tree, even when they reference other symbolic objects. As a result,
  the ``rebind()`` method cannot modify the value they are pointing to.

  For primitive types, ``pg.Ref()`` returns their values directly without
  creating a reference. For example, ``pg.Ref(1)`` and ``pg.Ref('abc')`` will
  simply return the values 1 and 'abc', respectively, without any additional
  referencing.
  """

  def __new__(cls, value: Any, **kwargs):
    del kwargs
    if isinstance(value, (bool, numbers.Number, str)):
      return value
    return object.__new__(cls)

  @object_utils.explicit_method_override
  def __init__(self, value: Any, **kwargs) -> None:
    super().__init__(**kwargs)
    if isinstance(value, Ref):
      value = value.value
    self._value = value

  def _on_parent_change(
      self,
      old_parent: Optional[base.Symbolic],
      new_parent: Optional[base.Symbolic]) -> None:
    if (new_parent is not None
        and isinstance(self._value, base.Symbolic)
        and self._value.sym_root is new_parent.sym_root):
      raise NotImplementedError('Self-referential object is not supported.')

  @property
  def value(self) -> Any:
    """Returns the referenced value."""
    return self._value

  def infer(self, **kwargs) -> Any:
    """Returns the referenced value."""
    return self._value

  def custom_apply(
      self,
      path: object_utils.KeyPath,
      value_spec: pg_typing.ValueSpec,
      allow_partial: bool = False,
      child_transform: Optional[Callable[
          [object_utils.KeyPath, pg_typing.Field, Any], Any]] = None
      ) -> Tuple[bool, Any]:
    """Validate candidates during value_spec binding time."""
    del child_transform
    # Check if the field being assigned could accept the referenced value.
    # We do not do any transformation, thus not passing the child transform.
    value_spec.apply(
        self._value,
        allow_partial=allow_partial)
    return (False, self)

  def _sym_clone(self, deep: bool, memo: Any = None) -> 'Ref':
    # Always create a new object.
    # TODO(daiyip): support deep clone with the update of reference when
    # the original value is updated.
    return Ref(self._value, allow_partial=self.allow_partial)

  def sym_eq(self, other: Any) -> bool:
    return isinstance(other, Ref) and self.value is other.value

  def sym_jsonify(self, **kwargs: Any) -> Any:
    raise TypeError(f'{self!r} cannot be serialized at the moment.')

  def __getstate__(self):
    raise TypeError(f'{self!r} cannot be pickled at the moment.')

  def format(
      self,
      compact: bool = False,
      verbose: bool = False,
      root_indent: int = 0,
      **kwargs: Any) -> str:
    value_str = object_utils.format(
        self._value,
        compact=compact, verbose=verbose, root_indent=root_indent + 1)
    if compact:
      return f'{self.__class__.__name__}({value_str})'
    else:
      return (f'{self.__class__.__name__}(\n'
              + '  ' * (root_indent + 1)
              + f'value = {value_str}\n'
              + '  ' * root_indent + ')')


def maybe_ref(value: Any) -> Optional[Ref]:
  """Returns a reference if a value is not symbolic or already has a parent."""
  if isinstance(value, base.Symbolic):
    if value.sym_parent is None:
      return value
  return Ref(value)



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for intercepting typing checking."""

import abc
from typing import Any, Callable, Optional, Tuple

from pyglove.core import object_utils
from pyglove.core.typing import class_schema


class CustomTyping(metaclass=abc.ABCMeta):
  """Interface of custom value type.

  Instances of subclasses of CustomTyping can be assigned to fields of
  any ValueSpec, and take over `apply` via `custom_apply` method.

  As a result, CustomTyping makes the schema system extensible without modifying
  existing value specs. For example, value generators can extend CustomTyping
  and be assignable to any fields.
  """

  @abc.abstractmethod
  def custom_apply(
      self,
      path: object_utils.KeyPath,
      value_spec: class_schema.ValueSpec,
      allow_partial: bool,
      child_transform: Optional[Callable[
          [object_utils.KeyPath, class_schema.Field, Any], Any]] = None
  ) -> Tuple[bool, Any]:
    """Custom apply on a value based on its original value spec.

    Args:
      path: KeyPath of current object under its object tree.
      value_spec: Original value spec for this field.
      allow_partial: Whether allow partial object to be created.
      child_transform: Function to transform child node values into their final
        values. Transform function is called on leaf nodes first, then on their
        parents, recursively.

    Returns:
      A tuple (proceed_with_standard_apply, value_to_proceed).
        If proceed_with_standard_apply is set to False, value_to_proceed
        will be used as final value.

    Raises:
      Error when the value is not compatible with the value spec.
    """
